

module fifo_ctrl (
 input rst,
 input clk,

 output reg inw_reject, // Asserted for a cycle when an IENW request cannot be granted due to lack of FIFO space
 input [4:0] wr_req,
 input [5*IO_W-1:0] wr_pack,
 output reg [4:0] wr_grant,
 input [4:0] wr_valid, // IENW & CNW interface

 input rd_req,
 output reg [IO_W-1:0] rd_pack,
 output reg   rd_valid,
 input [2:0] rd_fifon,
 input [NUM_VIA_W-1:0] rd_num_via, // send module interface

 input [WORD_W-1:0] mem_od,
 output [WORD_W-1:0] mem_id,
 output              mem_csel,
 output [MEM_ADD_W-1:0] mem_add,
 output mem_wr
);

   parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
   parameter MEM_ADD_W = 10;
   parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
	parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
   parameter NUM_VIA_W = FIFO_DEPTH_W; // Max number of via points in a packet = 2^NUM_VIA_W
	parameter REC_LIST_ADD = 5 * (2 ** FIFO_DEPTH_W);

   localparam IO_W = WORD_W/2;
	localparam HEADER_SIZE_W = (HEADER_SIZE == 3 ? 2 : (HEADER_SIZE < 8 ? 3 : (HEADER_SIZE < 16 ? 4 : 5))); // OK for HEADER_SIZE < 32
   localparam FIFO_NUM_PACK_W = FIFO_DEPTH_W-HEADER_SIZE_W;

   reg [2:0] wr_state;
	reg [2:0] next_wr_state;
	reg [2:0] wr_fifo_id;
	
   reg [FIFO_DEPTH_W*5-1:0] head;
	reg [FIFO_DEPTH_W*5-1:0] tail; // The each of them goes from 0 to (max_fifo_depth-1)
   reg [WORD_W+MEM_ADD_W:0] write_buff;
	reg [IO_W-1:0]    int_buff_wr;
	reg [FIFO_DEPTH_W-1:0] space_wr;
	wire [FIFO_DEPTH_W-1:0] tail_wr;
	wire [FIFO_DEPTH_W-1:0]	head_wr;
	wire [IO_W-1:0] sel_pack_wr;
	
   reg [HEADER_SIZE_W-1:0] pack_header_cnt;
	reg [NUM_VIA_W-1 :0] rd_via_cnt;
	reg [NUM_VIA_W-1 :0] old_num_via;
	reg [2:0] rd_state;
	reg [2:0] next_rd_state;
	reg [2:0] rd_fifo_id;
	reg [IO_W-1 :0] int_buff_rd;
	reg 		rd_mem_cs;
	reg [MEM_ADD_W-1:0] rd_mem_add;
	reg [3:0] rec_list_cnt;  // Counts 0 to 9, because writing happens once in two clock cycles

	wire [FIFO_DEPTH_W-1:0]	head_rd;

   reg rd_cycle;
	reg [(FIFO_NUM_PACK_W*5)-1 :0] fifo_num_pack;
   reg [WORD_W-1 :0] rd_mem_id;
	wire  rd_mem_wr;

 
   generate
      if (NUM_VIA_W >= IO_W)
         error_PackSize_gt_IOWidth error_inst0();
		if (FIFO_DEPTH_W > MEM_ADD_W-3)
		   error_MemDepth_lt_FifoNeed error_inst1();
		if (WORD_W-FIFO_DEPTH_W-1 <= 0)
		   error_FIFO_too_deep error_inst2();
	   if(HEADER_SIZE >= 32)
		   error_PackHeadSize_gte_32 error_inst3(); // Update the HEADER_SIZE_W computation to include this range, if necessary
		if(HEADER_SIZE <= 2)
			error_PackHeadSize_lte_2 error_inst4();
	endgenerate


//--------------------------------------------------------------------------------------//
//
//           FIFO write logic
//
//--------------------------------------------------------------------------------------//
// Write FSM
   always @(posedge clk or posedge rst)
      if(rst)
      begin
         wr_state <= 3'b000;
         wr_fifo_id <= 3'b000;
      end
      else
      begin
         wr_state <= next_wr_state;
			
         if(next_wr_state == 3'b100)
			begin
			   if(wr_fifo_id < 4)
               wr_fifo_id <= wr_fifo_id +1;
				else
				   wr_fifo_id <= 0;
			end
      end
		
	wire fifo_avail = (space_wr >= sel_pack_wr +HEADER_SIZE);
	
	always @(posedge clk or posedge rst)
	   if(rst)
		   inw_reject <= 1'b0;
		else
		   if(wr_state == 3'b000 && wr_fifo_id == 0 && wr_req[0] && !fifo_avail)
			   inw_reject <= 1'b1;
			else
			   inw_reject <= 1'b0;

   always @(*)
   begin
      case(wr_state)
         3'b000: next_wr_state = (wr_req[wr_fifo_id] && fifo_avail) ? 3'b001 : 3'b100;
         3'b001: next_wr_state = (wr_valid[wr_fifo_id]) ? 3'b011 : 3'b001;
         3'b010: next_wr_state = (wr_valid[wr_fifo_id]) ? 3'b011 : 3'b110;
         3'b011: next_wr_state = 3'b010;
         3'b100: next_wr_state = 3'b000;
			3'b110: next_wr_state = 3'b100;
        default: next_wr_state = 3'b100;
		endcase
   end

// Write internal variables

   assign tail_wr = tail[wr_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W];
	assign head_wr = head[wr_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W];
	assign sel_pack_wr = wr_pack[wr_fifo_id*IO_W+:IO_W];

// space_wr is used only in wr_state == 3'b000
// This is also the critical path
	reg [FIFO_DEPTH_W-1:0] space_wr_tmp;
	reg roll_over;

   always @(posedge clk or posedge rst)
	   if(rst)
		begin
		   space_wr_tmp <= {FIFO_DEPTH_W{1'b1}}; // DJ: Assuming full FIFO depth
			roll_over <= 1'b0;
		end
		else if(wr_state == 3'b100) // The state just before 000
		begin
		   space_wr_tmp <= head_wr - tail_wr - 1'd1;
			roll_over <= (tail_wr >= head_wr);
		end

   always @(*)
	   if(roll_over)
		   space_wr = space_wr_tmp + {1'b1, {FIFO_DEPTH_W{1'b0}}}; // DJ: Assuming full FIFO depth
		else
		   space_wr = space_wr_tmp;

/*
   always @(posedge clk or posedge rst)
	   if(rst)
		   space_wr <= {FIFO_DEPTH_W{1'b1}}; // DJ: Assuming full FIFO depth
		else if(wr_state == 3'b100) // The state just before 000
		begin
	      if(tail_wr >= head_wr)
		      space_wr <= head_wr + {1'b1, {FIFO_DEPTH_W{1'b0}}} - 1'd1 - tail_wr; // DJ: Assuming full FIFO depth
		   else
		      space_wr <= head_wr - tail_wr - 1'd1;
		end
*/

   always @(posedge clk or posedge rst)
	   if(rst)
		   tail <= {(FIFO_DEPTH_W*5){1'b0}};
		else if(wr_state == 3'b010)
		   tail[wr_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W] <= tail_wr + 1; // DJ: Allowing for roll-over: full FIFO depth
			
	always @(posedge clk or posedge rst)
	   if(rst)
		   int_buff_wr <= {IO_W{1'b0}}; // No need to reset in terms of functionality
		else if(next_wr_state == 3'b011)
		   int_buff_wr <= sel_pack_wr;
			
// Write memory interface
    always @(posedge clk or posedge rst)
	    if(rst)
		    write_buff <= {(WORD_W+MEM_ADD_W+1){1'b0}};
		 else if(wr_state == 3'b011)
		 begin
		    write_buff[WORD_W+MEM_ADD_W] <= 1'b1; // we can have a valid memory write in next cycle
			 write_buff[WORD_W+MEM_ADD_W-1:IO_W+MEM_ADD_W] <= sel_pack_wr;
			 write_buff[IO_W+MEM_ADD_W-1:MEM_ADD_W] <= int_buff_wr;
			 write_buff[MEM_ADD_W-1:0] <= tail_wr + {wr_fifo_id, {FIFO_DEPTH_W{1'b0}}}; // DJ: Check that unconnected bits are put to 0, full FIFO depth
		 end
		 else if(wr_state == 3'b110)
		 begin
		    write_buff[WORD_W+MEM_ADD_W] <= 1'b1;
		    write_buff[WORD_W+MEM_ADD_W-1 : FIFO_DEPTH_W+MEM_ADD_W+1] <= {(WORD_W-FIFO_DEPTH_W-1){1'b0}};  // MEM_ADD_W must not be too big
		    write_buff[FIFO_DEPTH_W+MEM_ADD_W] <= 1'b1;
			 write_buff[FIFO_DEPTH_W+MEM_ADD_W -1: MEM_ADD_W] <= head_wr;
		    write_buff[MEM_ADD_W-1:0] <= wr_fifo_id + REC_LIST_ADD;
		 end
		 else if(wr_state == 3'b000)
		    write_buff[WORD_W+MEM_ADD_W] <= 1'b0; // Finish the memory write
			 
// IENW/CNW interface
   always @(posedge clk or posedge rst)
	   if(rst)
		   wr_grant <= 5'b0_0000;
		else if(next_wr_state == 3'b001)
		   wr_grant[wr_fifo_id] <= 1'b1;
		else if(next_wr_state == 3'b011)
		   wr_grant[wr_fifo_id] <= 1'b0;


//--------------------------------------------------------------------------------------//
//
//           FIFO Read logic
//
//--------------------------------------------------------------------------------------//

// Read FSM
   always @(posedge clk or posedge rst)
	   if(rst)
		   rd_state <= 3'b111;
		else
		   rd_state <= next_rd_state;

   always @(*)
	   case(rd_state)
		   3'b000: next_rd_state = !rd_req ? 3'b000 : (rd_cycle ? 3'b101 : 3'b001);
			3'b001: next_rd_state = 3'b011;
			3'b010: next_rd_state = (rd_via_cnt == 0 && pack_header_cnt == 0) ? 3'b110 : 3'b011;
			3'b011: next_rd_state = 3'b010;
			3'b100: next_rd_state = 3'b000;
			3'b101: next_rd_state = 3'b001;
			3'b110: next_rd_state = 3'b100;
			3'b111: next_rd_state = (rec_list_cnt == 9) ? 3'b000 : 3'b111;
		  default: next_rd_state = 3'b000;
		endcase


// Read internal variables
   always @(posedge clk or posedge rst)
	   if(rst)
		   rec_list_cnt <= 4'b0000;
		else if(rd_state == 3'b111)
		   rec_list_cnt <= rec_list_cnt + 4'b0001;

	assign head_rd = head[rd_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W];

   always @(posedge clk or posedge rst)
	   if(rst)
		   head <= {FIFO_DEPTH_W*5{1'b0}};
		else if(rd_state == 3'b001)
		   head[rd_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W] <= head_rd+1;
		else if(rd_state == 3'b010)
		begin
		   if(rd_via_cnt == 0 && pack_header_cnt == 0) // Packet has been fully read out
			   head[rd_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W] <= head_rd+old_num_via; // DJ: Allowing for roll-over: full FIFO depth
			else
		      head[rd_fifo_id*FIFO_DEPTH_W+:FIFO_DEPTH_W] <= head_rd+1; // DJ: Allowing for roll-over: full FIFO depth
		end
		
	always @(posedge clk or posedge rst)
	   if(rst)
		begin
		   pack_header_cnt <= 0;
			rd_via_cnt <= 0;
			old_num_via <= 0;
			rd_fifo_id <= 0;
		end
		else if(rd_state == 3'b000 && rd_req)
		begin
		   pack_header_cnt <= HEADER_SIZE-1; // incorporating the reduction in state 001
			rd_via_cnt <= rd_num_via;
			rd_fifo_id <= rd_fifon;
		end
		else if(rd_state == 3'b010)
		begin
			if(pack_header_cnt > 0) pack_header_cnt <= pack_header_cnt -1'd1;
			else if(rd_via_cnt > 0)
			begin
				rd_via_cnt <= rd_via_cnt -1'd1;
				old_num_via <= old_num_via -1'd1;
			end
		end
		else if(rd_state == 3'b011 && pack_header_cnt == HEADER_SIZE-2) // Data for the 2nd word in the packet just got read; Assume HEADER_SIZE > 2.
		   old_num_via <= mem_od[NUM_VIA_W-1 :0];  // Assuming that the LSB's are the old number of vias. Also note: NUM_VIA_W lte IO_W

// Read memory interface

   always @(*)
      case(rd_state)
		   3'b001, 3'b011:	begin
				rd_mem_cs = 1'b1;
				rd_mem_add = head_rd + {rd_fifo_id, {FIFO_DEPTH_W{1'b0}}};
			end
			3'b010: begin
			   if(rd_via_cnt == 0 && pack_header_cnt == 0)
				   rd_mem_cs = 1'b0;
				else
				   rd_mem_cs = 1'b1;
					
				rd_mem_add = head_rd + {rd_fifo_id, {FIFO_DEPTH_W{1'b0}}};
			end
			3'b100, 3'b110: begin
			   rd_mem_cs = 1'b1;
				rd_mem_add = REC_LIST_ADD + rd_fifo_id;  // Updating the receive list
			end
			3'b111: begin
			   rd_mem_cs = 1'b1;
				rd_mem_add = REC_LIST_ADD + rec_list_cnt[3:1];
			end
			default: begin
			   rd_mem_cs = 1'b0;
				rd_mem_add = 0;
			end
		endcase

// Send-module interface
   always @(posedge clk or posedge rst)
      if(rst)
         int_buff_rd <= 0;
		else if(rd_state == 3'b011)
		   int_buff_rd <= mem_od[WORD_W-1 : IO_W];

		
	always @(*)
	   if(rd_state == 3'b011)
		begin
		   if(pack_header_cnt == HEADER_SIZE-2)  // LSB's of second word; Assumes HEADER_SIZE > 2
			begin
				rd_pack[NUM_VIA_W-1:0] = rd_via_cnt;
				rd_pack[IO_W-1 : NUM_VIA_W] = mem_od[IO_W-1 : NUM_VIA_W]; // Assumes IOW > NUM_VIA_W
			end
			else
				rd_pack = mem_od[IO_W-1 :0];
			rd_valid = 1'b1;
		end
		else if(rd_state == 3'b010)
		begin
		   rd_pack = int_buff_rd;
			rd_valid = 1'b1;
		end
		else
		begin
		   rd_pack = 0;
			rd_valid = 1'b0;
		end

   always @(posedge clk or posedge rst)
	   if(rst)
		   rd_mem_id <= 0;
		else if(rd_state == 3'b110)
		begin
		   rd_mem_id[FIFO_DEPTH_W-1 :0] <= head_rd;
			rd_mem_id[FIFO_DEPTH_W] <= fifo_num_pack[rd_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] > 1 ? 1'b1 : 1'b0; // It is >1 because
			          // fifo_num_pack is decremented in the same cycle. In case it aligns with the wr_state == 110 and for the same FIFO
						 // fifo_num_pack is not decremented, and the read FSM may end up wrongly choosing 0 from this statement.
						 // However, the write FSM will correct the mistake in the receive list in the following clock cycle.
		end
		else if(rd_state == 3'b111)
		begin
		   rd_mem_id[FIFO_DEPTH_W-1 :0] <= 0;
			rd_mem_id[FIFO_DEPTH_W] <= 0;  // Assuming that the first packet wouldn't finish writing within 10 clocks
		end
		else
		   rd_mem_id <= 0;

   assign rd_mem_wr = (rd_state == 3'b100) | (rd_state == 3'b111);

//------------------------------------------------------------
//
//  Keep count of number of packets in FIFO
//
//------------------------------------------------------------

   always @(posedge clk or posedge rst)
	   if(rst)
		   fifo_num_pack <= {(FIFO_NUM_PACK_W*5){1'b0}};
      else
		begin
         if(wr_state == 3'b110 && rd_state != 3'b110)
			   fifo_num_pack[wr_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] <= fifo_num_pack[wr_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] + 1;
         else if (wr_state != 3'b110 && rd_state == 3'b110)
			   fifo_num_pack[rd_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] <= fifo_num_pack[rd_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] - 1;
			else if (wr_state == 3'b110 && rd_state == 3'b110 && wr_fifo_id != rd_fifo_id)
			begin
			   fifo_num_pack[wr_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] <= fifo_num_pack[wr_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] + 1;
			   fifo_num_pack[rd_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] <= fifo_num_pack[rd_fifo_id*FIFO_NUM_PACK_W+:FIFO_NUM_PACK_W] - 1;
			end
      end

	
//--------------------------------------------------------------------------------------//
//
//           Memory interface
//
//--------------------------------------------------------------------------------------//

	always @(posedge clk or posedge rst)
	   if(rst)
		   rd_cycle <= 1'b0;
		else
		   rd_cycle <= !rd_cycle;

   assign mem_id = rd_cycle ? rd_mem_id : write_buff[WORD_W+MEM_ADD_W-1 :MEM_ADD_W];
	assign mem_add = rd_cycle ? rd_mem_add : write_buff[MEM_ADD_W-1 :0];
	assign mem_wr = rd_cycle ? rd_mem_wr : write_buff[WORD_W+MEM_ADD_W];
	assign mem_csel = rd_cycle ? rd_mem_cs : write_buff[WORD_W+MEM_ADD_W];


endmodule
