
module ie_node(
 input clk,
 input rst,
 input [ADD_SIZE-1:0] our_add,

 input    wr_req_ds,
 input [ADD_SIZE-1 :0] wr_add_ds, 
 input [IO_W-1:0] wr_pack_ds,
 output      wr_grant_ds,
 output      wr_reject_ds,
 input    wr_valid_ds,

 output reg   wr_req_us,
 output reg [ADD_SIZE-1 :0] wr_add_us, 
 output reg [IO_W-1:0] wr_pack_us,
 input        wr_grant_us,
 input		  wr_reject_us,
 output reg   wr_valid_us,

 output reg req_fifo_wr,
 output reg [IO_W-1:0] pack_fifo_wr,
 input      grant_fifo_wr,
 input      reject_fifo_wr,
 output reg valid_fifo_wr,
 
 input    req_fifo_rd,
 input [IO_W-1:0] pack_fifo_rd,
 output      grant_fifo_rd,
 input    valid_fifo_rd

    );

	parameter IO_W = 16;
	parameter FLOP_ON_IENW_REQ = 1;
	parameter FLOP_ON_IENW_GRANT_REJ = 1;
	parameter ADD_SIZE = 16;

   reg [2:0] state_wr, state_rd;
   reg [2:0] next_state_wr, next_state_rd;

   reg    wr_req_fw;
   reg [ADD_SIZE-1 :0] wr_add_fw; 
   reg [IO_W-1:0] wr_pack_fw;
   reg    wr_valid_fw;
	reg   wr_grant_fw;
	reg   reject_d, wr_reject_fw, grant_u_d, grant_f_d;

// State machines
 
   always @(posedge clk or posedge rst)
      if(rst)
		begin
  	      state_wr <= 3'b000;
			state_rd <= 3'b000;
			
			reject_d <= 1'b0;
			grant_f_d <= 1'b0;
			grant_u_d <= 1'b0;
		end
	   else
		begin
	      state_wr <= next_state_wr;
			state_rd <= next_state_rd;
			
			reject_d <= wr_reject_us;
			grant_f_d <= wr_grant_fw;
			grant_u_d <= wr_grant_us;
		end


   always @(*)
	begin
	   case(state_wr)
		   3'b000: next_state_wr = wr_req_ds ? (wr_add_ds == our_add ? 3'b011 : 3'b100) : 3'b000; // Idle state
			3'b001: next_state_wr = wr_req_ds ? 3'b001 : 3'b000; // Waiting for a rejected FIFO write to be aborted by upstream nodes
			3'b010: next_state_wr = wr_valid_ds ? 3'b010 : 3'b000;     // Burst data transmission to FIFO active
			3'b011: next_state_wr = reject_fifo_wr ? 3'b001 : (wr_valid_ds ? 3'b010 : 3'b011); // Waiting for a requested transmission to FIFO to start
			3'b100: next_state_wr = wr_req_ds | grant_f_d ? (wr_valid_ds ? 3'b101 : 3'b100) : 3'b000; // Waiting for a requested data forward transaction to start
			3'b101: next_state_wr = wr_valid_ds ? 3'b101 : 3'b000; // Burst data forward transaction is active
		  default: next_state_wr = 3'b000;
		 endcase
	end
	
	always @(*)
	begin
	   case(state_rd)
		   3'b000: next_state_rd = req_fifo_rd ? 3'b100 : (wr_req_fw ? 3'b001 : 3'b000); // Idle state
			3'b001: next_state_rd = wr_req_fw | grant_u_d ? (wr_valid_fw ? 3'b010 : 3'b001) : 3'b000; // Waiting for a requested data forwarding transaction to start
			3'b010: next_state_rd = wr_valid_fw ? 3'b010 : 3'b000; // Burst data forwarding transaction ongoing
			3'b100: next_state_rd = reject_d ? 3'b111 : (valid_fifo_rd ? 3'b101 : 3'b100); // Waiting for a requested data transfer from FIFO to start
			3'b101: next_state_rd = valid_fifo_rd ? 3'b101 : 3'b000; // Burst data transfer from FIFO is active
			3'b111: next_state_rd = wr_req_fw ? 3'b001 : 3'b100; // After rejection of FIFO read, 1 cycle back-off.
		  default: next_state_rd = 3'b000;
		endcase
	end



// The forwarding interface
// May or may not include a register
generate
if(FLOP_ON_IENW_REQ)
begin
   always @(posedge clk or posedge rst)
      if(rst)
      begin
	      wr_req_fw <= 0;
		   wr_pack_fw <= 0;
			wr_add_fw <= 0;
			wr_valid_fw <= 0;
      end
		else if(next_state_wr == 3'b100 || next_state_wr == 3'b101)
		begin
	      wr_req_fw <= wr_req_ds;
		   wr_pack_fw <= wr_pack_ds;
			wr_add_fw <= wr_add_ds;
			wr_valid_fw <= wr_valid_ds;
		end
		else
      begin
	      wr_req_fw <= 0;
		   wr_pack_fw <= 0;
			wr_add_fw <= 0;
			wr_valid_fw <= 0;
      end
end
else begin
   always @(*)
		if(next_state_wr == 3'b100 || next_state_wr == 3'b101)
		begin
	      wr_req_fw = wr_req_ds;
		   wr_pack_fw = wr_pack_ds;
			wr_add_fw = wr_add_ds;
			wr_valid_fw = wr_valid_ds;
		end
		else
      begin
	      wr_req_fw = 0;
		   wr_pack_fw = 0;
			wr_add_fw = 0;
			wr_valid_fw = 0;
      end
end
endgenerate

// FIFO Write interface
   always @(posedge clk or posedge rst)
	   if(rst)
		begin
         req_fifo_wr <= 1'b0;
         pack_fifo_wr <= 0;
         valid_fifo_wr <= 1'b0;
      end
		else if(next_state_wr == 3'b011 || next_state_wr == 3'b010)
		begin
		   req_fifo_wr <= wr_req_ds;
         pack_fifo_wr <= wr_pack_ds;
         valid_fifo_wr <= wr_valid_ds;
		end
		else
		begin
         req_fifo_wr <= 1'b0;
         pack_fifo_wr <= 0;
         valid_fifo_wr <= 1'b0;
      end

   assign wr_grant_ds = (state_wr == 3'b011) ? grant_fifo_wr : (state_wr == 3'b100 ? wr_grant_fw : 1'b0);
	assign wr_reject_ds = (state_wr == 3'b011) ? reject_fifo_wr : (state_wr == 3'b100 ? wr_reject_fw : 1'b0);

// Interface to downstream node
	always @(*)
	begin
	   if(next_state_rd == 3'b100 || next_state_rd == 3'b101)
		begin
		   wr_req_us = req_fifo_rd;
			wr_pack_us = pack_fifo_rd;
			wr_valid_us = valid_fifo_rd;
			wr_add_us = -1;  // Address of ejection node
		end
		else if(next_state_rd == 3'b001 || next_state_rd == 3'b010)
		begin
		   wr_req_us = wr_req_fw;
			wr_pack_us = wr_pack_fw;
			wr_valid_us = wr_valid_fw;
			wr_add_us = wr_add_fw;
		end
		else
		begin
		   wr_req_us = 0;
			wr_pack_us = 0;
			wr_valid_us = 0;
			wr_add_us = 0;
		end
	end

	assign grant_fifo_rd = (state_rd == 3'b100) ? wr_grant_us : 1'b0;
	
generate
	if(FLOP_ON_IENW_GRANT_REJ)
	   always @(posedge clk or posedge rst)
		if(rst)
		begin
		   wr_grant_fw <= 0;
			wr_reject_fw <= 0;
		end
		else
		begin
		   wr_grant_fw <= (state_rd == 3'b001) ? wr_grant_us : 1'b0;
			wr_reject_fw <= (state_rd == 3'b001) ? wr_reject_us : 1'b0;
		end
			
	else
	   always @(*)
		begin
		   wr_grant_fw = (state_rd == 3'b001) ? wr_grant_us : 1'b0;
			wr_reject_fw = (state_rd == 3'b001) ? wr_reject_us : 1'b0;
		end
endgenerate


endmodule
