
module proc_node(
 input rst,
 input clk,

 input  [4:0] wr_req_ds,
 input  [5*IO_W-1:0] wr_pack_ds,
 output [4:0] wr_grant_ds,
 input  [4:0] wr_valid_ds, // IENW & CNW interface - downstream ports, driven by an upstream proc_node
 
 output [4:0] wr_req_us,
 output [5*IO_W-1:0] wr_pack_us,
 input  [4:0] wr_grant_us,
 output [4:0] wr_valid_us, // IENW & CNW interface - upstream ports, to drive a downstream proc_node
 
 input  [MEM_ADD_W:0] cpu_add,
 input   cpu_wr,
 input   cpu_cs,
 output [WORD_W-1:0] cpu_data_i,
 input  [WORD_W-1:0] cpu_data_o // Processor interface
);

parameter TIMEOUT_W = 10;
parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
parameter MEM_ADD_W = 10;
parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
parameter PN_Y_COORD = 0;
parameter PN_X_COORD = 0;
parameter COORD_SIZE = 8;
parameter NUM_VIA_W = FIFO_DEPTH_W;
parameter REC_LIST_ADD = 5 * (2 ** FIFO_DEPTH_W);
parameter TIMEOUT_COUNT_PREC = 0; // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

localparam SEND_LIST_BASE = 2**MEM_ADD_W;
localparam IO_W = WORD_W/2;

 wire rd_req;
 wire [IO_W-1:0] rd_pack;
 wire   rd_valid;
 wire [2:0] rd_fifon;
 wire [NUM_VIA_W-1:0] rd_num_via;
/*
 wire [MEM_ADD_W:0] cpu_add;
 wire  cpu_wr;
 wire  cpu_cs;
 wire [WORD_W-1:0] cpu_data_i;
 wire [WORD_W-1:0] cpu_data_o; */
 
 wire [WORD_W-1:0] cpu_data_i_bram;
 wire [4+TIMEOUT_W+NUM_VIA_W:0] cpu_data_i_sendl;
 
 
 wire [WORD_W-1:0] mem_od;
 wire [WORD_W-1:0] mem_id;
 wire              mem_csel;
 wire [MEM_ADD_W-1:0] mem_add;
 wire mem_wr;

 reg sendl;

send_mod #(
   .TIMEOUT_W(TIMEOUT_W),
   .NUM_VIA_W(NUM_VIA_W),
	.IO_W(IO_W),
	.PN_Y_COORD(PN_Y_COORD),
	.PN_X_COORD(PN_X_COORD),
	.COORD_SIZE(COORD_SIZE),
	.TIMEOUT_COUNT_PREC(TIMEOUT_COUNT_PREC)

) send_module_inst (
 .rst(rst),
 .clk(clk),
 
 .wr_req(wr_req_us),
 .wr_pack(wr_pack_us),
 .wr_grant(wr_grant_us),
 .wr_valid(wr_valid_us), // IENW & CNW interface

 .rd_req(rd_req),
 .rd_pack(rd_pack),
 .rd_valid(rd_valid),
 .rd_fifon(rd_fifon),
 .rd_num_via(rd_num_via), // FIFO-controller interface

 .cpu_add(cpu_add[2:0]),
 .cpu_wr(cpu_wr),
 .cpu_cs(cpu_cs & (cpu_add >= SEND_LIST_BASE) & (cpu_add < SEND_LIST_BASE+6)),
 .cpu_data_i(cpu_data_o[4+TIMEOUT_W+NUM_VIA_W:0]),
 .cpu_data_o(cpu_data_i_sendl)  // Send-list memory-like interface
 // CPU data format: {sent_toggle, ready_toggle, timeout, route_dir, num_via}. sent_toggle ignored when writing into send-list
);



fifo_ctrl #(
   .WORD_W(WORD_W),
   .MEM_ADD_W(MEM_ADD_W),
   .HEADER_SIZE(HEADER_SIZE),
	.FIFO_DEPTH_W(FIFO_DEPTH_W),
   .NUM_VIA_W(NUM_VIA_W),
	.REC_LIST_ADD(REC_LIST_ADD)
) fifo_ctrl_inst (
 .rst(rst),
 .clk(clk),

 .wr_req(wr_req_ds),
 .wr_pack(wr_pack_ds),
 .wr_grant(wr_grant_ds),
 .wr_valid(wr_valid_ds), // IENW & CNW interface

 .rd_req(rd_req),
 .rd_pack(rd_pack),
 .rd_valid(rd_valid),
 .rd_fifon(rd_fifon),
 .rd_num_via(rd_num_via), // send module interface

 .mem_od(mem_od),
 .mem_id(mem_id),
 .mem_csel(mem_csel),
 .mem_add(mem_add),
 .mem_wr(mem_wr)
);

block_ram #(
   .RAM_WIDTH(WORD_W),
   .RAM_ADDR_BITS(MEM_ADD_W)
) block_ram_inst (
 .clockA(clk),
 .clockB(clk),
 
 .odataA(mem_od),
 .addA(mem_add),
 .idataA(mem_id),
 .enableA(mem_csel),
 .write_enableA(mem_wr),

 .odataB(cpu_data_i_bram),
 .addB(cpu_add[MEM_ADD_W-1:0]),
 .idataB(cpu_data_o),
 .enableB(cpu_cs & (cpu_add < SEND_LIST_BASE)),
 .write_enableB(cpu_wr)
);

always @(posedge clk or posedge rst)
   if(rst)
	   sendl <= 1'b0;
	else if(cpu_add >= SEND_LIST_BASE)
	   sendl <= 1'b1;
	else
	   sendl <= 1'b0;

assign cpu_data_i = sendl ? cpu_data_i_sendl : cpu_data_i_bram;

/*
processor #(
 .MEM_ADD_W(MEM_ADD_W),
 .WORD_W(WORD_W)
) processor_inst(
 .rst(rst),
 .clk(clk),

 .cpu_add(cpu_add),
 .cpu_wr(cpu_wr),
 .cpu_cs(cpu_cs),
 .cpu_data_i(cpu_data_i),
 .cpu_data_o(cpu_data_o)
); */

endmodule
