
module proc_node_with_ie(
 input rst,
 input clk,

 input  ie_req_ds,
 input  [IO_W-1:0] ie_pack_ds,
 input  [(COORD_SIZE*2)-1:0] ie_add_ds,
 output ie_grant_ds,
 output ie_reject_ds,
 input  ie_valid_ds, // IENW interface - downstream ports, driven by an upstream proc_node
 
 output ie_req_us,
 output [IO_W-1:0] ie_pack_us,
 output [(COORD_SIZE*2)-1:0] ie_add_us,
 input  ie_grant_us,
 input  ie_reject_us,
 output ie_valid_us, // IENW interface - upstream ports, to drive a downstream proc_node

 input  [3:0] wr_req_ds,
 input  [4*IO_W-1:0] wr_pack_ds,
 output [3:0] wr_grant_ds,
 input  [3:0] wr_valid_ds, // CNW interface - downstream ports, driven by an upstream proc_node
 
 output [3:0] wr_req_us,
 output [4*IO_W-1:0] wr_pack_us,
 input  [3:0] wr_grant_us,
 output [3:0] wr_valid_us, // CNW interface - upstream ports, to drive a downstream proc_node
 
 input  [MEM_ADD_W:0] cpu_add,
 input   cpu_wr,
 input   cpu_cs,
 output [WORD_W-1:0] cpu_data_i,
 input  [WORD_W-1:0] cpu_data_o // Processor interface
);

parameter TIMEOUT_W = 10;
parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
parameter MEM_ADD_W = 10;
parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
parameter PN_Y_COORD = 0;
parameter PN_X_COORD = 0;
parameter COORD_SIZE = 8;
parameter NUM_VIA_W = FIFO_DEPTH_W;
parameter REC_LIST_ADD = 5 * (2 ** FIFO_DEPTH_W);
parameter FLOP_ON_IENW_REQ = 0;
parameter FLOP_ON_IENW_GRANT_REJ = 0;
parameter TIMEOUT_COUNT_PREC = 0; // Timeout is counted every 2**TIMEOUT_COUNT_PREC clocks

localparam SEND_LIST_BASE = 2**MEM_ADD_W;
localparam IO_W = WORD_W/2;

 wire rd_req;
 wire [IO_W-1:0] rd_pack;
 wire   rd_valid;
 wire [2:0] rd_fifon;
 wire [NUM_VIA_W-1:0] rd_num_via;

 
 wire [WORD_W-1:0] cpu_data_i_bram;
 wire [4+TIMEOUT_W+NUM_VIA_W:0] cpu_data_i_sendl;
 
 
 wire [WORD_W-1:0] mem_od;
 wire [WORD_W-1:0] mem_id;
 wire              mem_csel;
 wire [MEM_ADD_W-1:0] mem_add;
 wire mem_wr;

 reg sendl;
 wire inj_reject_local;
 
 wire   req_fifo_wr;
 wire [IO_W-1:0] pack_fifo_wr;
 wire      grant_fifo_wr;
 wire   valid_fifo_wr;
 
 wire    req_fifo_rd;
 wire [IO_W-1:0] pack_fifo_rd;
 wire      grant_fifo_rd;
 wire    valid_fifo_rd;
 
 wire [(COORD_SIZE*2) -1:0] our_add;

ie_node #(
  .IO_W(IO_W),
  .FLOP_ON_IENW_REQ(FLOP_ON_IENW_REQ),
  .FLOP_ON_IENW_GRANT_REJ(FLOP_ON_IENW_GRANT_REJ),
  .ADD_SIZE(2*COORD_SIZE)
)
 ie_node_inst(
 .clk(clk),
 .rst(rst),
 .our_add(our_add),

 .wr_req_ds(ie_req_ds),
 .wr_add_ds(ie_add_ds), 
 .wr_pack_ds(ie_pack_ds),
 .wr_grant_ds(ie_grant_ds),
 .wr_reject_ds(ie_reject_ds),
 .wr_valid_ds(ie_valid_ds),

 .wr_req_us(ie_req_us),
 .wr_add_us(ie_add_us),
 .wr_pack_us(ie_pack_us),
 .wr_grant_us(ie_grant_us),
 .wr_reject_us(ie_reject_us),
 .wr_valid_us(ie_valid_us),

 .req_fifo_wr(req_fifo_wr),
 .pack_fifo_wr(pack_fifo_wr),
 .grant_fifo_wr(grant_fifo_wr),
 .valid_fifo_wr(valid_fifo_wr),
 .reject_fifo_wr(inj_reject_local),
 
 .req_fifo_rd(req_fifo_rd),
 .pack_fifo_rd(pack_fifo_rd),
 .grant_fifo_rd(grant_fifo_rd),
 .valid_fifo_rd(valid_fifo_rd)

    );

send_mod #(
   .TIMEOUT_W(TIMEOUT_W),
   .NUM_VIA_W(NUM_VIA_W),
	.IO_W(IO_W),
	.PN_Y_COORD(PN_Y_COORD),
	.PN_X_COORD(PN_X_COORD),
	.COORD_SIZE(COORD_SIZE),
	.TIMEOUT_COUNT_PREC(TIMEOUT_COUNT_PREC)

) send_module_inst (
 .rst(rst),
 .clk(clk),
 .our_add(our_add),
 
 .wr_req({wr_req_us, req_fifo_rd}),
 .wr_pack({wr_pack_us, pack_fifo_rd}),
 .wr_grant({wr_grant_us, grant_fifo_rd}),
 .wr_valid({wr_valid_us, valid_fifo_rd}), // IENW & CNW interface

 .rd_req(rd_req),
 .rd_pack(rd_pack),
 .rd_valid(rd_valid),
 .rd_fifon(rd_fifon),
 .rd_num_via(rd_num_via), // FIFO-controller interface

 .cpu_add(cpu_add[2:0]),
 .cpu_wr(cpu_wr),
 .cpu_cs(cpu_cs & (cpu_add >= SEND_LIST_BASE) & (cpu_add < SEND_LIST_BASE+6)),
 .cpu_data_i(cpu_data_o[4+TIMEOUT_W+NUM_VIA_W:0]),
 .cpu_data_o(cpu_data_i_sendl)  // Send-list memory-like interface
 // CPU data format: {sent_toggle, ready_toggle, timeout, route_dir, num_via}. sent_toggle ignored when writing into send-list
);



fifo_ctrl #(
   .WORD_W(WORD_W),
   .MEM_ADD_W(MEM_ADD_W),
   .HEADER_SIZE(HEADER_SIZE),
	.FIFO_DEPTH_W(FIFO_DEPTH_W),
   .NUM_VIA_W(NUM_VIA_W),
	.REC_LIST_ADD(REC_LIST_ADD)
) fifo_ctrl_inst (
 .rst(rst),
 .clk(clk),

 .wr_req({wr_req_ds, req_fifo_wr}),
 .wr_pack({wr_pack_ds, pack_fifo_wr}),
 .wr_grant({wr_grant_ds, grant_fifo_wr}),
 .wr_valid({wr_valid_ds, valid_fifo_wr}), // IENW & CNW interface
 .inw_reject(inj_reject_local),

 .rd_req(rd_req),
 .rd_pack(rd_pack),
 .rd_valid(rd_valid),
 .rd_fifon(rd_fifon),
 .rd_num_via(rd_num_via), // send module interface

 .mem_od(mem_od),
 .mem_id(mem_id),
 .mem_csel(mem_csel),
 .mem_add(mem_add),
 .mem_wr(mem_wr)
);

block_ram #(
   .RAM_WIDTH(WORD_W),
   .RAM_ADDR_BITS(MEM_ADD_W)
) block_ram_inst (
 .clockA(clk),
 .clockB(clk),
 
 .odataA(mem_od),
 .addA(mem_add),
 .idataA(mem_id),
 .enableA(mem_csel),
 .write_enableA(mem_wr),

 .odataB(cpu_data_i_bram),
 .addB(cpu_add[MEM_ADD_W-1:0]),
 .idataB(cpu_data_o),
 .enableB(cpu_cs & (cpu_add < SEND_LIST_BASE)),
 .write_enableB(cpu_wr)
);

always @(posedge clk or posedge rst)
   if(rst)
	   sendl <= 1'b0;
	else if(cpu_add >= SEND_LIST_BASE)
	   sendl <= 1'b1;
	else
	   sendl <= 1'b0;

assign cpu_data_i = sendl ? cpu_data_i_sendl : cpu_data_i_bram;


endmodule
