`timescale 1ns / 1ps
/*
Description:
1. Assert req to write a 10-via packet into FIFO2, a 5 via-packet to FIFO3 and a 4 via packet to FIFO4
2. Write them in sequence of getting grant
3. As soon as the 10-via packet is finished writing put into send-list with 10 via points with routing direction 1
4. As soon as send module asserts req to output in direction 1, give it grant
5. This should ensure some overlap between reading and writing into different FIFO's
6. After the writing of 2nd and 3rd packets are over, update the send-list to send them out with number of vias reduced by 1, but
  both to direction 3.
7. When the send-module asserts the output req in direction 3, give grant. Should happen 2 times for the 2 packets.
8. At the end, the CPU reads all receive list and send list entries


Things verified:
1. FIFO2 reading and FIFO4 writing are concurrent
2. Data integrity of packets - number of vias read, updation of number of vias in packet, routing to correct direction
3. At one point two FIFO's (3 & 4) requested for the same output interface (#3). Correctly handled.
4. After a successful packet send, the receive-list is updated in the same cycle as the sent_toggle is toggled.
4. At the end, the receive list is empty but with correct head pointers
5. At the end, the send-list has ready_toggle == sent_toggle, with 0 for FIFO's 0 and 1, and 1 for FIFO's 2, 3, 4.
*/

module test_bench1(

    );

   parameter TIMEOUT_W = 10;
   parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
   parameter MEM_ADD_W = 10;
   parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
	parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
   parameter NUM_VIA_W = FIFO_DEPTH_W; // Max number of via points in a packet = 2^NUM_VIA_W

   localparam IO_W = WORD_W/2;

   reg clk;
	reg rst;

 reg [4:0] wr_req;
 reg [5*IO_W-1:0] wr_pack;
 wire [4:0] wr_grant;
 reg [4:0] wr_valid; // IENW & CNW interface

 wire [4:0] wr_req_op;
 wire [5*IO_W-1:0] wr_pack_op;
 reg [4:0] wr_grant_op;
 wire [4:0] wr_valid_op; // IENW & CNW interface

 reg  [MEM_ADD_W:0] cpu_add;
 reg   cpu_wr;
 reg   cpu_cs;
 wire [WORD_W-1:0] cpu_data_i;
 wire [WORD_W-1:0] cpu_data_o;

 reg ready_toggle;
 reg [TIMEOUT_W-1:0] timeout;
 reg [2:0] route_dir;
 reg [NUM_VIA_W-1:0] num_via;


   initial
	begin
	   clk <= 1'b0;
      forever
	      #5 clk <= !clk;
   end
	
	initial
	begin
	   rst <= 1'b1;
		wr_req <= 5'b11100;
		wr_pack <= 80'h0004_0005_000a_0000_0000;
		wr_valid <= 0;
		
		#20 rst <= 1'b0;

// Write a packet with 10 via points into FIFO2
			 
		@(posedge wr_grant[2]);
		@(posedge clk);
		wr_req[2] <= 0;

		@(posedge clk);
		@(posedge clk);
		@(posedge clk);
		wr_pack[47:32] <= 16'h550a;
		wr_valid[2] <= 1;
		
		#300 wr_valid[2] <= 0;
		
// Write a packet with 5 via points into FIFO3

		@(posedge wr_grant[3]);
		@(posedge clk);
		wr_req[3] <= 0;

		@(posedge clk);
		wr_pack[63:48] <= 16'haa05;
		wr_valid[3] <= 1;
		
		#200 wr_valid[3] <= 0;
		
// Write a packet with 4 via points into FIFO4

		@(posedge wr_grant[4]);
		@(posedge clk);
		wr_req[4] <= 0;

		@(posedge clk);
		@(posedge clk);
		wr_pack[79:64] <= 16'hff04;
		wr_valid[4] <= 1;
		
		#180 wr_valid[4] <= 0;
				
	end

initial
begin
   cpu_add <= 11'b010_1000_0010; // FIFO2 entry in receive list
	cpu_wr <= 0;
	cpu_cs <= 1;
	
	@(cpu_data_i);
	@(cpu_data_i);

	cpu_add <= 11'b100_0000_0010; // send-list entry for FIFO2
	cpu_wr <= 1;
	ready_toggle <= 1;
	timeout <= 20;
	route_dir <= 1;
	num_via <= 10;
	
	@(posedge clk)
	cpu_wr <= 0;
	cpu_add <= 11'b010_1000_0100; // FIFO4 entry in receive list
	
	@(posedge clk);
	@(posedge clk);
	@(cpu_data_i);

	cpu_add <= 11'b100_0000_0011; // send-list entry for FIFO3
	cpu_wr <= 1;
	ready_toggle <= 1;
	timeout <= 20;
	route_dir <= 3;
	num_via <= 4;
	
	@(posedge clk)
	
	cpu_add <= 11'b100_0000_0100; // send-list entry for FIFO4
	cpu_wr <= 1;
	ready_toggle <= 1;
	timeout <= 20;
	route_dir <= 3;
	num_via <= 3;
	
	@(posedge clk)
	cpu_wr <= 0;
	cpu_add <= 11'b010_1000_0100; // FIFO4 entry in receive list

	@(posedge clk)
	@(posedge clk)
	@(cpu_data_i)
// Simulation finishing now
	cpu_add <= 11'b010_1000_0011;
	
	@(posedge clk)
	cpu_add <= 11'b010_1000_0010;
	
	@(posedge clk)
	cpu_add <= 11'b010_1000_0001;
	
	@(posedge clk)
	cpu_add <= 11'b010_1000_0000;

// Send-list
	@(posedge clk)
	cpu_add <= 11'b100_0000_0000;
	
	@(posedge clk)
	cpu_add <= 11'b100_0000_0001;
	
	@(posedge clk)
	cpu_add <= 11'b100_0000_0010;
	
	@(posedge clk)
	cpu_add <= 11'b100_0000_0011;
	
	@(posedge clk)
	cpu_add <= 11'b100_0000_0100;

#100 $finish;
end

initial
begin
	wr_grant_op <= 0;

   @(posedge clk);
	@(wr_req_op);
	@(posedge clk);
	
	wr_grant_op[1] <= 1;
	
	@(wr_valid_op);
	@(posedge clk);
	wr_grant_op <= 0;
	
	@(wr_req_op);
	@(posedge clk);
	
	wr_grant_op[3] <= 1;
	
	@(wr_valid_op);
	@(posedge clk);
	wr_grant_op <= 0;
	
	@(wr_req_op);
	@(posedge clk);
	
	wr_grant_op[3] <= 1;
	
	@(wr_valid_op);
	@(posedge clk);
	wr_grant_op <= 0;
end

assign cpu_data_o[3+NUM_VIA_W+TIMEOUT_W] = ready_toggle;
assign cpu_data_o[NUM_VIA_W+2+TIMEOUT_W:NUM_VIA_W+3] = timeout;
assign cpu_data_o[NUM_VIA_W+2:NUM_VIA_W] = route_dir;
assign cpu_data_o[NUM_VIA_W-1:0]	= num_via;

proc_node proc_node_inst(
 .rst(rst),
 .clk(clk),

 .wr_req_ds(wr_req),
 .wr_pack_ds(wr_pack),
 .wr_grant_ds(wr_grant),
 .wr_valid_ds(wr_valid), // IENW & CNW interface - downstream ports, driven by an upstream proc_node
 
 .wr_req_us(wr_req_op),
 .wr_pack_us(wr_pack_op),
 .wr_grant_us(wr_grant_op),
 .wr_valid_us(wr_valid_op), // IENW & CNW interface - upstream ports, to drive a downstream proc_node
 
 .cpu_add(cpu_add),
 .cpu_wr(cpu_wr),
 .cpu_cs(cpu_cs),
 .cpu_data_i(cpu_data_i),
 .cpu_data_o(cpu_data_o) // Processor interface
);

endmodule

