`timescale 1ns / 1ps
/*
Description:
1. A packet with 5 via points is written into FIFO 0
2. When the writing is visible from CPU through receive list, the CPU updates the corresponding sendlist entry
 with a timeout of 20, and routing direction of 2
3. No grant is ever provided to the output interface

Things verified:
1. CPU sees the correct entry in receive list
2. Send-module sees the correct entries in ready-toggle, timeout, new number of vias and take action accordingly, i.e. assert
 the correct output req with correct new number of vias. The req direction changes to 0 after timeout.

*/

module test_bench3(

    );

   parameter TIMEOUT_W = 10;
   parameter WORD_W = 32;    // Width of each word, same as width of the memory interface. Must be even.
   parameter MEM_ADD_W = 10;
   parameter HEADER_SIZE = 5; // size of ftr packet header in number of words
	parameter FIFO_DEPTH_W = 7; // Depth of FIFO in number of words = (2^FIFO_DEPTH_W -1) : -1 for distinguishing empty/full
   parameter NUM_VIA_W = FIFO_DEPTH_W; // Max number of via points in a packet = 2^NUM_VIA_W

   localparam IO_W = WORD_W/2;

   reg clk;
	reg rst;

 reg [4:0] wr_req;
 reg [5*IO_W-1:0] wr_pack;
 wire [4:0] wr_grant;
 reg [4:0] wr_valid; // IENW & CNW interface

 wire [4:0] wr_req_op;
 wire [5*IO_W-1:0] wr_pack_op;
 reg [4:0] wr_grant_op;
 wire [4:0] wr_valid_op; // IENW & CNW interface

 reg  [MEM_ADD_W:0] cpu_add;
 reg   cpu_wr;
 reg   cpu_cs;
 wire [WORD_W-1:0] cpu_data_i;
 wire [WORD_W-1:0] cpu_data_o;

 reg ready_toggle;
 reg [TIMEOUT_W-1:0] timeout;
 reg [2:0] route_dir;
 reg [NUM_VIA_W-1:0] num_via;


   initial
	begin
	   clk <= 1'b0;
      forever
	      #5 clk <= !clk;
   end
	
	initial
	begin
	   rst <= 1'b1;
		wr_req <= 0;
		wr_pack <= 0;
		wr_valid <= 0;
		wr_grant_op <= 0;
		
		#20 rst <= 1'b0;

// Write a packet with 5 via points
		#20 wr_pack[15:0] <= 5;
		    wr_req[0] <= 1;
			 
		@(posedge wr_grant[0]);
		@(posedge clk);
		wr_req <= 0;
		
		@(posedge clk);
		@(posedge clk);
		@(posedge clk);
		wr_pack[15:0] <= 16'h5505;
		wr_valid[0] <= 1;
		
		#200 wr_valid <= 0;

					 
		#1500 $finish;
		
	end

initial
begin
   cpu_add <= 11'b010_1000_0000;
	cpu_wr <= 0;
	cpu_cs <= 1;
	
	@(cpu_data_i);
	@(cpu_data_i);
	@(posedge clk);
	cpu_add <= 11'b100_0000_0000; // send-list
	cpu_wr <= 1;
	cpu_cs <= 1;
	ready_toggle <= 1;
	timeout <= 20;
	route_dir <= 2;
	num_via <= 4;
	
	@(posedge clk)
	cpu_wr <= 0;
	cpu_add <= 11'b010_1000_0000;
	
end

assign cpu_data_o[WORD_W-1 :4+NUM_VIA_W+TIMEOUT_W] = 0;
assign cpu_data_o[3+NUM_VIA_W+TIMEOUT_W] = ready_toggle;
assign cpu_data_o[NUM_VIA_W+2+TIMEOUT_W:NUM_VIA_W+3] = timeout;
assign cpu_data_o[NUM_VIA_W+2:NUM_VIA_W] = route_dir;
assign cpu_data_o[NUM_VIA_W-1:0]	= num_via;

proc_node #(.TIMEOUT_COUNT_PREC(1))
 proc_node_inst(
 .rst(rst),
 .clk(clk),

 .wr_req_ds(wr_req),
 .wr_pack_ds(wr_pack),
 .wr_grant_ds(wr_grant),
 .wr_valid_ds(wr_valid), // IENW & CNW interface - downstream ports, driven by an upstream proc_node
 
 .wr_req_us(wr_req_op),
 .wr_pack_us(wr_pack_op),
 .wr_grant_us(wr_grant_op),
 .wr_valid_us(wr_valid_op), // IENW & CNW interface - upstream ports, to drive a downstream proc_node
 
 .cpu_add(cpu_add),
 .cpu_wr(cpu_wr),
 .cpu_cs(cpu_cs),
 .cpu_data_i(cpu_data_i),
 .cpu_data_o(cpu_data_o) // Processor interface
);

endmodule
