`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
/*
Description:
1. First there is an incoming request on IENW to have a packet forwarded to the next node.
2. Even before a grant is received for this request, a request is received also from the send-module interface.
3. After a while a grant is received on the IENW (to be associated with the first request), which is to be forwarded to the upstream node.
4. After a while, the upstream node starts sends a packet, which should be forwarded to the downstream node.
5. After this packet send is finished, a new request should be raised by the present node towards the downstream node with respect to the already-asserted request received from the send-module interface.
6. A new grant is received with respect to the last request, and data is sent out from the send-module.
7. While this transfer is active, the upstream node makes a new request to write data into the present node's FIFO through the FIFO-controller. The present node should be able to forward this request to the FIFO-controller, even when making a transfer from the send-module interface.
8. The FIFO-controller gives grant, and a corresponding packet transfer is performed.  Right after its completion, the upstream node makes a fresh request for another packet to be forwarded. However, the send-module interface was still making its transfer into the IENW. So, the new request should be blocked by the present node.
9. After a while the send-module transfer is completed. The present node should be able to forward the request from its upstream node to its downstream node in the next clock cycle.

Things verified:
1. Both state machines come back to the '000' state at the end of a packet transfer, which is signalled by at least 1 cycle of valid = 0.
2. The node forwards a virtual circuit setup request to the next node, if it is not destined for itself, and then does the transfer accordingly.
3. When a forwarding circuit is setup, the Send-module cannot access the bus, and vice-versa.
4. It is possible to read and write the FIFO's of the same node at the same time, i.e. the node can process an incoming IENW request with the same destination address as that of the present node, and a bus access request from the send-module.
*/
//////////////////////////////////////////////////////////////////////////////////
module test_bench5(
    );
parameter IO_W = 16;

   reg clk;
	reg rst;

   initial
	begin
	   clk <= 1'b0;
      forever
	      #5 clk <= !clk;
   end

reg wr_req_ds, req_fifo_rd, wr_grant_us, grant_fifo_wr;
reg [15:0] wr_add_ds;
reg [IO_W-1:0] wr_pack_ds, pack_fifo_rd;
reg   wr_valid_ds, valid_fifo_rd;

initial
begin
 rst <= 1;
 wr_req_ds <= 0;
 req_fifo_rd <= 0;
 wr_grant_us <= 0;
 grant_fifo_wr <= 0;

 #20 rst <= 0;
 wr_add_ds <= 16'h10_0A;
 wr_pack_ds <= 6;
 wr_valid_ds <= 0;
 
 #10;
// @(posedge clk);
 @(posedge clk)
    wr_req_ds <= 1;

 #45;
 pack_fifo_rd <= 10;
 valid_fifo_rd <= 0;
 
 @(posedge clk)
    req_fifo_rd <= 1;

  #105 wr_grant_us <= 1;
  
 #50 wr_req_ds <= 0;
     wr_pack_ds <= 0;
	  
 #50 wr_valid_ds <= 1;
     wr_pack_ds <= 16'hf0_0f;
	  
 #10 wr_grant_us <= 0;
 
 #90 wr_valid_ds <= 0;
 
 #20  wr_pack_ds <= 0;
 
 #45 wr_grant_us <= 1;
 
 #30;
 @(posedge clk);
     req_fifo_rd <= 0;
	  
 #20
     valid_fifo_rd <= 1;
	  pack_fifo_rd <= 16'ha245;

 #50 wr_grant_us <= 0;
 
 #25 wr_req_ds <= 1;
    wr_pack_ds <= 8;
	 wr_add_ds <= 16'h05_02;
 
 #30 grant_fifo_wr <= 1;
 
 #40 wr_req_ds <= 0;
 
 #20
    wr_add_ds <= 0;
	 wr_pack_ds <= 16'h7654;
	 wr_valid_ds <= 1;
	 
 #80 grant_fifo_wr <= 0;
 
// #20 valid_fifo_rd <= 0;
 
 #100 wr_valid_ds <= 0;
 
 #20 wr_req_ds <= 1;
    wr_add_ds <= 45;
	 wr_pack_ds <= 13;
 
 #60 valid_fifo_rd <= 0;
 
 #100 $finish;
end

ie_node #(
  .IO_W(16),
  .FLOP_ON_IENW_REQ(1),
  .FLOP_ON_IENW_GRANT_REJ(1)
) ie_node_inst(
 .clk(clk),
 .rst(rst),
 .our_add(16'h05_02),

 .wr_req_ds(wr_req_ds),
 .wr_add_ds(wr_add_ds), 
 .wr_pack_ds(wr_pack_ds),
 .wr_grant_ds(),
 .wr_reject_ds(),
 .wr_valid_ds(wr_valid_ds),

 .wr_req_us(),
 .wr_add_us(), 
 .wr_pack_us(),
 .wr_grant_us(wr_grant_us),
 .wr_reject_us(1'b0),
 .wr_valid_us(),

 .req_fifo_wr(),
 .pack_fifo_wr(),
 .grant_fifo_wr(grant_fifo_wr),
 .reject_fifo_wr(1'b0),
 .valid_fifo_wr(),
 
 .req_fifo_rd(req_fifo_rd),
 .pack_fifo_rd(pack_fifo_rd),
 .grant_fifo_rd(),
 .valid_fifo_rd(valid_fifo_rd)

    );


endmodule
