`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
/*
Description:
1. First there is an incoming request on IENW to have a packet forwarded to the next node.
2. Even before a grant is received for this request, a request is received also from the send-module interface.
3. After a while a grant is received on the IENW (to be associated with the first request), which is to be forwarded to the upstream node.
4. After a while, the upstream node starts sends a packet, which should be forwarded to the downstream node.
5. After this packet send is finished, a new request should be raised by the present node towards the downstream node with respect to the already-asserted request received from the send-module interface.
6. A new grant is received with respect to the last request, and data is sent out from the send-module.
7. While this transfer is active, the upstream node makes a new request to write data into the present node's FIFO through the FIFO-controller. The present node should be able to forward this request to the FIFO-controller, even when making a transfer from the send-module interface.
8. The FIFO-controller gives grant, and a corresponding packet transfer is performed.  Right after its completion, the upstream node makes a fresh request for another packet to be forwarded. However, the send-module interface was still making its transfer into the IENW. So, the new request should be blocked by the present node.
9. After a while the send-module transfer is completed. The present node should be able to forward the request from its upstream node to its downstream node in the next clock cycle.

Things verified:
1. Both state machines come back to the '000' state at the end of a packet transfer, which is signalled by at least 1 cycle of valid = 0.
2. The node forwards a virtual circuit setup request to the next node, if it is not destined for itself, and then does the transfer accordingly.
3. When a forwarding circuit is setup, the Send-module cannot access the bus, and vice-versa.
4. It is possible to read and write the FIFO's of the same node at the same time, i.e. the node can process an incoming IENW request with the same destination address as that of the present node, and a bus access request from the send-module.
*/
//////////////////////////////////////////////////////////////////////////////////
module test_bench6(
    );
parameter IO_W = 16;

   reg clk;
	reg rst;

   initial
	begin
	   clk <= 1'b0;
      forever
	      #5 clk <= !clk;
   end

wire wr_req_us, wr_reject_ds, req_fifo_wr1;

reg wr_req_ds, wr_grant_us, wr_reject_us;
reg [15:0] wr_add_ds;
reg [IO_W-1:0] wr_pack_ds;
reg   wr_valid_ds;

wire wr_req01, wr_valid01, wr_grant01, wr_reject01;
wire [15:0] wr_add01;
wire [IO_W-1 :0] wr_pack01;

wire wr_req12, wr_valid12, wr_grant12, wr_reject12;
wire [15:0] wr_add12;
wire [IO_W-1 :0] wr_pack12;

reg grant_fifo_wr0, reject_fifo_wr0;
reg grant_fifo_wr1, reject_fifo_wr1;
reg grant_fifo_wr2, reject_fifo_wr2;

reg req_fifo_rd0, valid_fifo_rd0;
reg [IO_W-1 :0] pack_fifo_rd0;
reg req_fifo_rd1, valid_fifo_rd1;
reg [IO_W-1 :0] pack_fifo_rd1;
reg req_fifo_rd2, valid_fifo_rd2;
reg [IO_W-1 :0] pack_fifo_rd2;

initial
begin
	wr_grant_us <= 0;
	wr_reject_us <= 0;

   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
   @(posedge wr_req_us);
	#40;
	@(posedge clk)
	   wr_reject_us <= 1;	
	@(posedge clk)
	   wr_reject_us <= 0;
	
	#100 $finish;

end

initial
begin
   grant_fifo_wr1 <= 0;
	reject_fifo_wr1 <= 0;
	
	@(posedge req_fifo_wr1);
	#100;
	@(posedge clk)
	   reject_fifo_wr1 <= 1;
	@(posedge clk)
	   reject_fifo_wr1 <= 0;
end

initial
begin
   req_fifo_rd0 <= 0;
	req_fifo_rd1 <= 0;
	req_fifo_rd2 <= 0;
	rst <= 1;
	wr_req_ds <= 0;
	
	#20 rst <= 0;
	
	#40
	wr_req_ds <= 1;
	wr_add_ds <= 16'h03_03;
	wr_pack_ds <= 15;
	wr_valid_ds <= 0;
// At this point a forwarding request is active through all the nodes

	#20 req_fifo_rd0 <= 1; req_fifo_rd1 <= 1; req_fifo_rd2 <= 1;
	    pack_fifo_rd0 <= 3; pack_fifo_rd1 <= 4; pack_fifo_rd2 <= 1;
		 valid_fifo_rd0 <= 0; valid_fifo_rd1 <= 0; valid_fifo_rd2 <= 0;
		 
// After a while the active request will be rejected.
	@(posedge wr_reject_ds);
	@(posedge clk)
		wr_req_ds <= 0;
		
	@(posedge clk)
	   wr_req_ds <= 1;
		
// After the second rejection of a forward request, let's try to write to the middle node.
	@(posedge wr_reject_ds);
	@(posedge clk)
		wr_req_ds <= 0;
	wr_add_ds <= 16'h05_01;
	wr_pack_ds <= 13;
		
	@(posedge clk)
	   wr_req_ds <= 1;

// After rejection of FIFO write, let's go back to writing output, but with a delay before deasserting rejected request

	@(posedge wr_reject_ds);
	#100;
	@(posedge clk)
		wr_req_ds <= 0;
	wr_add_ds <= 16'h05_0a;
	wr_pack_ds <= 10;
		
	@(posedge clk)
	   wr_req_ds <= 1;

end


ie_node #(
  .IO_W(16),
  .FLOP_ON_IENW_REQ(0),
  .FLOP_ON_IENW_GRANT_REJ(0)
) ie_node_inst0(
 .clk(clk),
 .rst(rst),
 .our_add(16'h05_00),

 .wr_req_ds(wr_req_ds),
 .wr_add_ds(wr_add_ds), 
 .wr_pack_ds(wr_pack_ds),
 .wr_grant_ds(),
 .wr_reject_ds(wr_reject_ds),
 .wr_valid_ds(wr_valid_ds),

 .wr_req_us(wr_req01),
 .wr_add_us(wr_add01), 
 .wr_pack_us(wr_pack01),
 .wr_grant_us(wr_grant01),
 .wr_reject_us(wr_reject01),
 .wr_valid_us(wr_valid01),

 .req_fifo_wr(),
 .pack_fifo_wr(),
 .grant_fifo_wr(grant_fifo_wr0),
 .reject_fifo_wr(reject_fifo_wr0),
 .valid_fifo_wr(),
 
 .req_fifo_rd(req_fifo_rd0),
 .pack_fifo_rd(pack_fifo_rd0),
 .grant_fifo_rd(),
 .valid_fifo_rd(valid_fifo_rd0)

    );

ie_node #(
  .IO_W(16),
  .FLOP_ON_IENW_REQ(0),
  .FLOP_ON_IENW_GRANT_REJ(0)
) ie_node_inst1(
 .clk(clk),
 .rst(rst),
 .our_add(16'h05_01),

 .wr_req_ds(wr_req01),
 .wr_add_ds(wr_add01), 
 .wr_pack_ds(wr_pack01),
 .wr_grant_ds(wr_grant01),
 .wr_reject_ds(wr_reject01),
 .wr_valid_ds(wr_valid01),

 .wr_req_us(wr_req12),
 .wr_add_us(wr_add12), 
 .wr_pack_us(wr_pack12),
 .wr_grant_us(wr_grant12),
 .wr_reject_us(wr_reject12),
 .wr_valid_us(wr_valid12),

 .req_fifo_wr(req_fifo_wr1),
 .pack_fifo_wr(),
 .grant_fifo_wr(grant_fifo_wr1),
 .reject_fifo_wr(reject_fifo_wr1),
 .valid_fifo_wr(),
 
 .req_fifo_rd(req_fifo_rd1),
 .pack_fifo_rd(pack_fifo_rd1),
 .grant_fifo_rd(),
 .valid_fifo_rd(valid_fifo_rd1)

    );

ie_node #(
  .IO_W(16),
  .FLOP_ON_IENW_REQ(1),
  .FLOP_ON_IENW_GRANT_REJ(1)
) ie_node_inst2(
 .clk(clk),
 .rst(rst),
 .our_add(16'h05_02),

 .wr_req_ds(wr_req12),
 .wr_add_ds(wr_add12), 
 .wr_pack_ds(wr_pack12),
 .wr_grant_ds(wr_grant12),
 .wr_reject_ds(wr_reject12),
 .wr_valid_ds(wr_valid12),

 .wr_req_us(wr_req_us),
 .wr_add_us(), 
 .wr_pack_us(),
 .wr_grant_us(wr_grant_us),
 .wr_reject_us(wr_reject_us),
 .wr_valid_us(),

 .req_fifo_wr(),
 .pack_fifo_wr(),
 .grant_fifo_wr(grant_fifo_wr2),
 .reject_fifo_wr(reject_fifo_wr2),
 .valid_fifo_wr(),
 
 .req_fifo_rd(req_fifo_rd2),
 .pack_fifo_rd(pack_fifo_rd2),
 .grant_fifo_rd(),
 .valid_fifo_rd(valid_fifo_rd2)

    );

endmodule
